----------------------------------------------------------------------------------
-- Company: ETH Zurich
-- Engineer: Pascal Engeler <engeler.pascal@gmail.com>
-- 
-- Create Date:    14:32:01 09/06/2024 
-- Design Name: 
-- Module Name:    DistanceUnwrapper - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
use IEEE.NUMERIC_STD.ALL;

entity DistanceUnwrapper is
    Port (  clk100			: in STD_LOGIC;
			en 				: in  STD_LOGIC;
			dr				: in STD_LOGIC;
		    distance_in	 	: in  STD_LOGIC_VECTOR(20 downto 0);
            distance_out    : out STD_LOGIC_VECTOR(24 downto 0);
			dr_out			: out STD_LOGIC);
end DistanceUnwrapper;

architecture Behavioral of DistanceUnwrapper is

type STATE_T is (IDLE_S, STATE1, STATE2, STATE3, STATE4, STATE5, STATE6);

signal current_state : STATE_T := IDLE_S;

signal currentDistance : STD_LOGIC_VECTOR(24 downto 0) := (others => '0');
signal prevDistance : STD_LOGIC_VECTOR(24 downto 0) := (others => '0');
signal latchedDistance : STD_LOGIC_VECTOR(24 downto 0) := (others => '0');
signal distanceInWide : STD_LOGIC_VECTOR(24 downto 0) := (others => '0');
signal calcDelta : STD_LOGIC_VECTOR(24 downto 0) := (others => '0');

signal maxDelta : STD_LOGIC_VECTOR(24 downto 0) := "0000100000000000000000000";
signal corrDelta : STD_LOGIC_VECTOR(24 downto 0) := "0001000000000000000000000";
signal corrDeltaN : STD_LOGIC_VECTOR(24 downto 0) := "1111000000000000000000000";
signal currentOffset : STD_LOGIC_VECTOR(24 downto 0) := (others => '0');

signal newIsLarger : STD_LOGIC := '0';

signal dr_prev : STD_LOGIC := '0';

begin

distanceInWide(20 downto 0) <= distance_in;
distanceInWide(24 downto 21) <= (others => '0');

distance_out <= currentDistance;


dr_update: process(clk100)
begin
	if rising_edge(clk100) then
		dr_prev <= dr;
	end if;
end process;



proc: process(clk100)
begin
	if rising_edge(clk100) then
		if en = '1' then
			case current_state is
			when IDLE_S =>
				if dr = '1' and dr_prev = '0' then
					current_state <= STATE1;
					latchedDistance <= distanceInWide;
					--calculate relevant delta distance here
					if prevDistance > distanceInWide then
						newIsLarger <= '0';
						calcDelta <= prevDistance - distanceInWide;
					else 
						newIsLarger <= '1';
						calcDelta <= distanceInWide - prevDistance;
					end if;
				else
					current_state <= IDLE_S;
					latchedDistance <= latchedDistance;
					newIsLarger <= newIsLarger;
					calcDelta <= calcDelta;
				end if;
				currentOffset <= currentOffset;
				currentDistance <= currentDistance;
				dr_out <= '0';
				prevDistance <= prevDistance;

			when STATE1 =>
				current_state <= STATE2;
				--calculate change to offset here 
				if calcDelta > maxDelta then
					--we have wrapped around 
					if newIsLarger then
						--went negative
						currentOffset <= currentOffset + corrDeltaN;
					else 
						--went more positive
						currentOffset <= currentOffset + corrDelta;
					end if;
				else 
					currentOffset <= currentOffset;
				end if;
				latchedDistance <= latchedDistance;
				newIsLarger <= newIsLarger;
				calcDelta <= calcDelta;
				currentDistance <= currentDistance;
				dr_out <= dr_out;
				prevDistance <= prevDistance;
					
			when STATE2 => 
			--calculate output distance, set next prevDistance here
				current_state <= STATE3;
				currentDistance <= currentOffset + latchedDistance;
				latchedDistance <= latchedDistance;
				newIsLarger <= newIsLarger;
				calcDelta <= calcDelta;
				currentOffset <= currentOffset;
				dr_out <= dr_out;
				prevDistance <= latchedDistance;

			when STATE3 =>
			--set output ready here	
				current_state <= IDLE_S;
				dr_out <= '1';
				latchedDistance <= latchedDistance;
				newIsLarger <= newIsLarger;
				calcDelta <= calcDelta;
				currentOffset <= currentOffset;
				currentDistance <= currentDistance;
				prevDistance <= prevDistance;
			end case;

		else 
			currentDistance <= (others => '0');
			prevDistance <= (others => '0');
			current_state <= IDLE_S;
			latchedDistance <= (others => '0');
			newIsLarger <= '0';
			calcDelta <= (others => '0');
			currentOffset <= (others => '0');
			dr_out <= '0';
		end if;

	end if;
		
end process;

end Behavioral;

